#!/usr/bin/env python

"""
Program data structures. There are two separate kinds of structures: those with
context, which are the values manipulated by programs, and those without
context, which are typically constant things which are stored alongside the
program but which are wrapped in context-dependent structures in the running
program.

Copyright (C) 2007, 2008, 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The principal value structure class in this module is the Attr class which
represents attributes of objects and retains the context of each reference to an
attribute. This class models program behaviour at run-time.

The central data structure classes in this module are the following:

  * Class
  * Function
  * Module

All of the above support the Naming interface either explicitly or through
general conformance, meaning that all can be asked to provide their 'full_name'
using the method of that name.

Additionally, all of the above also support a dictionary interface in order to
access names within their defined scopes. Specific methods also exist in order
to distinguish between certain kinds of attributes:

  * Class: class_attributes, all_class_attributes, instance_attributes, all_attributes
  * Function: parameters, locals, all_locals
  * Module: module_attributes

These specific methods are useful in certain situations.

The above classes also provide an 'astnode' attribute, indicating the AST node
where each such object is defined.
"""

from micropython.program import DataObject, DataValue, ReplaceableContext, PlaceholderContext
from micropython.common import AtLeast, InspectError

def shortrepr(obj):
    if obj is None:
        return repr(None)
    else:
        return obj.__shortrepr__()

# Mix-ins and abstract classes.

class Naming:

    "A mix-in providing naming conveniences."

    def full_name(self):
        if self.name is not None:
            return self.parent.full_name() + "." + self.name
        else:
            return self.parent.full_name()

class NamespaceDict:

    "A mix-in providing dictionary methods."

    def __init__(self, module=None):
        self.module = module
        self.namespace = {}
        self.globals = set()
        self.finalised = 0

        # Attributes accessed on objects, potentially narrowing their types.
        # Specific namespaces should define known names during initialisation.
        # For example, functions can define names belonging to parameters.

        # Attribute users, defining names which use attributes.

        self.attribute_users = [{}] # stack of assignments
        self.user_shelves = []
        self.loop_users = [{}]      # stack of loop nodes

        # Scope usage, indicating the origin of names.

        self.scope_usage = [{}]     # stack of scope usage
        self.scope_shelves = []

        # Define attribute usage to identify active program sections.
        # Attribute users are AST nodes defining names.

        self.all_attribute_users = set()

    # Attribute/name definition and access.

    def __delitem__(self, name):
        del self.namespace[name]

    def has_key(self, name):
        return self.namespace.has_key(name)

    def keys(self):
        return self.namespace.keys()

    def values(self):
        return self.namespace.values()

    def items(self):
        return self.namespace.items()

    def __getitem__(self, name):
        return self.namespace[name]

    def get_using_node(self, name, node):

        """
        Access the given 'name' through this namespace, making use of the module
        and builtins namespaces if necessary, annotating the given 'node' with
        the scope involved.
        """

        attr, scope, full_name = self._get_with_scope(name)

        if scope is not None:
            node._scope = scope
            self.note_scope(name, scope)

        if full_name is not None:
            self.use_specific_attribute(full_name, name)

        return attr

    def _get_with_scope(self, name, external=0):

        module = self.module
        builtins = module and module.builtins or None
        importer = module and module.importer or None

        # Constants.

        if importer is not None and importer.predefined_constants.has_key(name):
            return importer.get_predefined_constant(name), "constant", None

        # Locals.

        elif not external and self.has_key(name):
            return self[name], "local", self.full_name()

        # Globals.

        elif module is not None and module is not self and module.has_key(name):
            return module[name], "global", module.full_name()

        # Builtins.

        elif builtins is not None and builtins.has_key(name):
            return builtins[name], "builtins", builtins.full_name()

        # Unknown.

        else:
            return None, None, None

    def get(self, name, default=None):
        return self.namespace.get(name, default)

    def __setitem__(self, name, value):
        self.set(name, value)

    def set(self, name, value, single_assignment=1):

        """
        A more powerful set operation, making 'name' refer to 'value' whilst
        indicating whether a 'single_assignment' (true by default) occurs in
        this operation (or whether the operation covers potentially many
        assignments in the lifetime of a program).
        """

        if name in self.globals:
            self.module.set(name, value, 0)
        else:
            self._set(name, value, single_assignment)
            self.note_scope(name, "local")

    def set_module(self, name, value):

        """
        A specialised set operation, making 'name' refer to 'value' in the
        context of making a module reference available in association with
        'name' as part of the import of that module or a submodule of that
        module.
        """

        self._set(name, value, 1)

    def _set(self, name, attr_or_value, single_assignment=1):

        """
        The underlying set operation associating 'name' with the given
        'attr_or_value'.
        See: docs/assignment.txt
        """

        # Add and/or obtain the namespace entry.

        if not self.namespace.has_key(name):
            self.namespace[name] = Attr(None, self, name)

        attr = self.namespace[name]
        self._set_using_attr(attr, attr_or_value, single_assignment)

    def _set_using_attr(self, attr, attr_or_value, single_assignment=1):

        # Handle attribute assignment as well as assignment of basic objects.
        # Attempt to fix the context if not explicitly defined.

        if isinstance(attr_or_value, Attr):
            context_values = self.get_updated_context_values(attr_or_value.context_values)
        else:
            context_values = self.get_updated_context_values([self.get_context_and_value(attr_or_value)])

        attr.update(context_values, single_assignment)

    def get_context_and_value(self, value):

        "Return a context, value tuple for the given 'value'."

        # Functions have a replaceable context.

        if isinstance(value, Function):
            return (ReplaceableContext, value)

        # Classes use placeholder contexts which cannot be replaced but which
        # do not communicate useful contextual information.

        elif isinstance(value, Class):
            return (PlaceholderContext, value)

        # Other values employ themselves as the context.

        else:
            return (value, value)

    def get_updated_context_values(self, context_values):

        """
        Adapt the contexts found in the given 'context_values', returning a new
        set.
        See: docs/assignment.txt
        """

        results = set()

        for context, value in context_values:

            # Set the context of instances to themselves.

            if isinstance(value, Instance):
                results.add((value, value))
            else:
                results.add((context, value))

        return results

    def make_global(self, name):

        "Declare 'name' as a global in the current namespace."

        if not self.namespace.has_key(name):
            self.globals.add(name)
            self.note_scope(name, "global")
            return 1
        else:
            return 0

    # Attribute positioning.

    def attributes_as_list(self):

        "Return the attributes in a list."

        self.finalise_attributes()
        l = [None] * len(self.keys())
        for attr in self.values():
            l[attr.position] = attr
        return l

    def finalise_attributes(self):

        "Make sure all attributes are fully defined."

        if self.finalised:
            return

        # The default action is to assign attribute positions sequentially.

        for i, attr in enumerate(self.values()):
            attr.position = i

        self.finalised = 1

    def unfinalise_attributes(self):

        "Open attribute definitions to editing and subsequent finalisation."

        self.finalised = 0

    # Attribute usage methods.

    def finalise_attribute_usage(self):

        "Propagate attribute usage for the namespace to the importer."

        module = self.module
        importer = module and module.importer

        if importer is not None:
            for names in self.get_all_attribute_usage():
                importer.use_names(names, self.full_name())

    def get_all_attribute_usage(self):

        """
        Return a set of all usage tuples for attributes used with the different
        local names.
        """

        usage = set()
        for user in self.all_attribute_users:
            for name, attrnames in user._attrnames.items():
                usage.add(tuple(attrnames))
        return usage

    def use_attribute(self, name, attrname):

        "Declare the usage on 'name' of the given 'attrname'."

        return self._use_attribute(name, attrname)

    def use_specific_attribute(self, objname, attrname):

        """
        Note attribute usage specifically on 'objname' - an object which is
        known at inspection time - or in the current unit if 'objname' is None,
        nominating a specific attribute 'attrname'.

        This bypasses attribute user mechanisms.
        """

        from_name = self.full_name()
        objname = objname or from_name
        module = self.module
        importer = module and module.importer

        if importer is not None:
            importer.use_specific_name(objname, attrname, from_name)

    # These shadow various methods in the InspectedModule class, and provide
    # implementations generally.

    def _use_attribute(self, name, attrname):

        """
        Indicate the use of the given 'name' in this namespace of an attribute
        with the given 'attrname'.
        """

        for users in (self.attribute_users[-1], self.loop_users[-1]):

            # Add the usage to all current users.

            if users.has_key(name):
                for user in users[name]:
                    user._attrnames[name].add(attrname)

        users = self.attribute_users[-1]

        if users.has_key(name):
            return users[name]
        else:
            return []

    def _define_attribute_user(self, node):

        """
        Define 'node' as the user of attributes, indicating the point where the
        user is defined.
        """

        name = node.name
        self._define_attribute_user_for_name(node, name)

    def _define_attribute_user_for_name(self, node, name):

        "Define 'node' as the user of attributes for the given 'name'."

        users = self.attribute_users[-1]

        # This node overrides previous definitions.

        users[name] = set([node])

        # Record the attribute combinations for the name.

        self._init_attribute_user_for_name(node, name)

        # Propagate any loop usage forward to any redefinition of a name.

        loop_users = self.loop_users[-1]

        if loop_users.has_key(name):
            for loop_user in loop_users[name]:
                node._attrnames[name].update(loop_user._attrnames[name])
            del loop_users[name]

        # Remember this user.

        self.all_attribute_users.add(node)

    def _init_attribute_user_for_name(self, node, name):

        "Make sure that 'node' is initialised for 'name'."

        if not hasattr(node, "_attrnames"):
            node._attrnames = {}

        node._attrnames[name] = set()

    # Branch management methods.

    def _new_branchpoint(self):

        """
        Establish a new branchpoint where several control-flow branches diverge
        and subsequently converge.
        """

        self.user_shelves.append([])
        self.scope_shelves.append([])

    def _new_branch(self, loop_node=None):

        """
        Establish a new control-flow branch, transferring attribute usage to
        the new branch so that it may be augmented for each name locally.

        If the optional 'loop_node' is given, add it as an active user to be
        informed of attribute usage.
        """

        # Retain a record of active users.

        new_users = {}
        new_users.update(self.attribute_users[-1])
        self.attribute_users.append(new_users)

        # Where a loop is the cause of the branch, register the loop node as a
        # user of each name so that attribute usage is also recorded for the
        # loop.

        loop_users = {}
        loop_users.update(self.loop_users[-1])
        self.loop_users.append(loop_users)

        if loop_node is not None:
            for name in new_users.keys():
                if not loop_users.has_key(name):
                    loop_users[name] = set([loop_node])
                else:
                    loop_users[name] = loop_users[name].union([loop_node])
                self._init_attribute_user_for_name(loop_node, name)

        # Retain a record of scope usage.

        scope_usage = {}
        scope_usage.update(self.scope_usage[-1])
        self.scope_usage.append(scope_usage)

    def _abandon_branch(self):
        pass

    def _shelve_branch(self):

        """
        Shelve the current control-flow branch, recording the attribute usage
        for subsequent merging. If this branch should be abandoned, the usage
        observations are still recorded but will not contribute to subsequent
        observations after a merge.
        """

        users = self.attribute_users.pop()
        self.user_shelves[-1].append(users)

        scope_usage = self.scope_usage.pop()
        self.scope_shelves[-1].append(scope_usage)

    def _merge_branches(self):

        """
        Merge control-flow branches. This should find the users active within
        each branch, which have been "shelved", and update the active users
        dictionary with these contributions.
        """

        # Combine the attribute users. This ensures that a list of users
        # affected by attribute usage is maintained for the current branch.

        users = self.attribute_users[-1]
        new_users = {}

        all_shelved_users = self.user_shelves.pop()
        all_user_names = set()

        # Find all the names defined by the branches.

        for shelved_users in all_shelved_users:
            all_user_names.update(shelved_users.keys())

        # Copy all definitions from the branches for the names, maintaining
        # the existing users where a branch does not redefine a name.

        for shelved_users in all_shelved_users:
            for name in all_user_names:

                if shelved_users.has_key(name):
                    nodes = shelved_users[name]
                else:
                    nodes = users.get(name, set())

                if nodes:
                    if not new_users.has_key(name):
                        new_users[name] = set(nodes)
                    else:
                        new_users[name].update(nodes)

        self.attribute_users[-1] = new_users

        # Combine the scope usage.

        scope_usage = self.scope_usage[-1]
        new_scope_usage = {}

        all_scope_usage = self.scope_shelves.pop()
        all_scope_names = set()

        # Find all the names for whom scope information has been defined.

        for shelved_usage in all_scope_usage:
            all_scope_names.update(shelved_usage.keys())

        for shelved_usage in all_scope_usage:
            for name in all_scope_names:

                # Find the recorded scope for the name.

                if shelved_usage.has_key(name):
                    scope = shelved_usage[name]
                elif scope_usage.has_key(name):
                    scope = scope_usage[name]

                # If no scope is recorded, find a suitable external source.

                else:
                    attr, scope, full_name = self._get_with_scope(name, external=1)

                # Attempt to record the scope, testing for conflicts.

                if scope:
                    if not new_scope_usage.has_key(name):
                        new_scope_usage[name] = scope
                    elif new_scope_usage[name] != scope:
                        raise InspectError("Scope conflict for %r: defined as both %s and %s." % (name, new_scope_usage[name], scope))

    # Scope usage methods.

    def note_scope(self, name, scope):

        "Note usage of 'name' from the given 'scope' in the current namespace."

        self.scope_usage[-1][name] = scope

    def used_in_scope(self, name, scope):

        """
        Return whether 'name' is used from the given 'scope' in the current
        namespace.
        """

        scope_usage = self.scope_usage[-1]
        return scope_usage.get(name) == scope

# Program data structures.

class Attr:

    "An attribute entry having a context."

    def __init__(self, position, parent, name):

        """
        Initialise the attribute with the given 'position' within the collection
        of attributes of its 'parent', indicating its 'name'.
        """

        self.position = position
        self.parent = parent
        self.name = name

        # Possible values.

        self.context_values = set()

        # Number of assignments per name.

        self.assignments = None

    # Value-related methods.

    def get_contexts(self):
        return [c for (c, v) in self.context_values]

    def get_values(self):
        return [v for (c, v) in self.context_values]

    def get_context(self):
        if len(self.context_values) == 1:
            return self.get_contexts()[0]
        else:
            return None

    def get_value(self):
        if len(self.context_values) == 1:
            return self.get_values()[0]
        else:
            return None

    def update(self, context_values, single_assignment):

        """
        Update the attribute, adding the 'context_values' provided to the
        known details associated with the attribute, changing the number of
        assignments according to the 'single_assignment' status of the
        operation, where a true value indicates that only one assignment is
        associated with the update, and a false value indicates that potentially
        many assignments may be involved.
        """

        if self.assignments is None:
            if single_assignment:
                self.assignments = 1
            else:
                self.assignments = AtLeast(1)
        else:
            if single_assignment:
                self.assignments += 1
            else:
                self.assignments += AtLeast(1)

        self.context_values.update(context_values)

    def is_constant(self):

        """
        Return whether this attribute references something that can be regarded
        as being constant within a particular scope.
        """

        return self.assignments == 1

    def is_strict_constant(self):

        """
        Return whether this attribute references something that can be regarded
        as being constant.
        """

        value = self.get_value()
        return not (value is None or isinstance(value, Instance))

    def is_static_attribute(self):

        """
        Return whether this attribute is defined on a fixed/static object such
        as a class or a module.
        """

        return isinstance(self.parent, (Class, Module))

    def defines_ambiguous_class(self):

        "Return whether this attribute defines more than one class."

        if self.assignments > 1:
            have_class = 0
            for obj in self.get_values():
                if isinstance(obj, Class):
                    if have_class:
                        return 1
                    have_class = 1

        return 0

    def defined_within_hierarchy(self):

        """
        Return whether the parent and context of the attribute belong to the
        same class hierarchy.
        """

        # Must be defined within a class.

        if isinstance(self.parent, Class):

            # To be sure, all contexts must be classes and be the same as the
            # parent, or be a superclass of the parent, or be a subclass of the
            # parent.

            for context in self.get_contexts():
                if not (
                    isinstance(context, Class) and (
                    context is self.parent or
                    context.has_subclass(self.parent) or
                    self.parent.has_subclass(context))
                    ):
                    return 0

            return 1

        # Instance attributes are not defined within a hierarchy.

        else:
            return 0

    def defined_outside_hierarchy(self):

        """
        Return whether the parent and context of the attribute never belong to
        the same class hierarchy.
        """

        # Must be defined within a class.

        if isinstance(self.parent, Class):

            # To be sure, all contexts must be classes and be the same as the
            # parent, or be a superclass of the parent, or be a subclass of the
            # parent.

            for context in self.get_contexts():
                if not (
                    isinstance(context, Class) and not (
                    context is self.parent or
                    context.has_subclass(self.parent) or
                    self.parent.has_subclass(context))
                    ):
                    return 0

            return 1

        # Instance attributes are not defined within a hierarchy.

        else:
            return 0

    def __repr__(self):
        return "Attr(%r, %s, %r) # {[%s] (%r)}" % (
            self.position, shortrepr(self.parent), self.name,
            self._context_values_str(), self.assignments
            )

    def __shortrepr__(self):
        return "Attr(%r, %s, %r)" % (
            self.position, shortrepr(self.parent), self.name
            )

    def _context_values_str(self):
        l = []
        for (c, v) in self.context_values:
            l.append("(c=%s, v=%s)" % (shortrepr(c), shortrepr(v)))
        return ", ".join(l)

# Instances are special in that they need to be wrapped together with context in
# a running program, but they are not generally constant.

class Instance:

    "A placeholder indicating the involvement of an instance."

    def __init__(self):
        self.parent = None

        # Image generation details.

        self.location = None

    def __repr__(self):
        return "Instance()"

    __shortrepr__ = __repr__

class Constant:

    "A superclass for all constant or context-free structures."

    pass

# Data objects appearing in programs before run-time.

class Const(Constant, Instance):

    "A constant object with no context."

    def __init__(self, value):
        Instance.__init__(self)
        self.value = value

    def get_value(self):
        return self.value

    def __repr__(self):
        if self.location is not None:
            return "Const(%r, location=%r)" % (self.value, self.location)
        else:
            return "Const(%r)" % self.value

    __shortrepr__ = __repr__

    # Support constants as dictionary keys in order to build constant tables.

    def __eq__(self, other):
        return other is not None and self.value == other.value and self.value.__class__ is other.value.__class__

    def __hash__(self):
        return hash(self.value)

    def value_type_name(self):
        return "__builtins__." + self.value.__class__.__name__

class Class(NamespaceDict, Naming, Constant):

    "An inspected class."

    def __init__(self, name, parent, module=None, node=None):

        """
        Initialise the class with the given 'name', 'parent' object, optional
        'module' and optional AST 'node'.
        """

        NamespaceDict.__init__(self, module)
        self.name = name
        self.parent = parent
        self.astnode = node
        node._def = self

        # Superclasses, descendants and attributes.

        self.bases = []
        self.descendants = set()
        self.instattr = set()           # instance attributes
        self.relocated = set()          # attributes which do not have the same
                                        # position as those of the same name in
                                        # some superclasses

        # Caches.

        self.reset_caches()

        # Image generation details.

        self.location = None
        self.code_location = None
        self.code_body_location = None  # corresponds to the instantiator

        self.instantiator = None
        self.instance_template_location = None # for creating instances at run-time

        # Program-related details.

        self.blocks = None
        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

        # Add this class to its attributes.

        self.set("__class__", self)

    def reset_caches(self):

        "Reset the caches."

        self.all_instattr = None        # cache for instance_attributes
        self.all_instattr_names = None  # from all_instattr
        self.all_classattr = None       # cache for all_class_attributes
        self.all_classattr_names = None # from all_classattr
        self.allattr = None             # cache for all_attributes
        self.allattr_names = None       # from allattr

    def __repr__(self):
        if self.location is not None:
            return "Class(%r, %s, location=%r)" % (self.name, shortrepr(self.parent), self.location)
        else:
            return "Class(%r, %s)" % (self.name, shortrepr(self.parent))

    def __shortrepr__(self):
        return "Class(%r, %s)" % (self.name, shortrepr(self.parent))

    def get_body_block(self):
        return self.get_instantiator().blocks[0]

    # Namespace-related methods.

    def get_updated_context_values(self, context_values):

        """
        Adapt the contexts found in the given 'context_values', returning a new
        set.
        See: docs/assignment.txt
        """

        results = set()

        for context, value in context_values:

            # Change the ownership of functions.

            if context is ReplaceableContext and value is not None and isinstance(value, Function):
                results.add((self, value))
            else:
                results.add((context, value))

        return NamespaceDict.get_updated_context_values(self, results)

    def finalise_attributes(self):

        "Make sure that all attributes are fully defined."

        if self.finalised:
            return

        self.finalise_class_attributes()
        self.finalise_instance_attributes()
        self.finalised = 1

    def unfinalise_attributes(self):

        "Open attribute definitions to editing and subsequent finalisation."

        self.reset_caches()
        self.finalised = 0

    # Convenience methods for accessing functions and methods.

    def get_instantiator(self):

        "Return a function which can be used to instantiate the class."

        if self.instantiator is None:
            self.instantiator = self.get_init_method().as_instantiator()
        return self.instantiator

    def get_init_method(self):
        return self.all_class_attributes()["__init__"].get_value()

    # Class-specific methods.

    def add_base(self, base):
        self.bases.append(base)
        base.add_descendant(self)

    def add_instance_attribute(self, name):
        self.instattr.add(name)

    def add_descendant(self, cls):
        self.descendants.add(cls)
        for base in self.bases:
            base.add_descendant(cls)

    def has_subclass(self, other):
        return other in self.descendants

    def all_descendants(self):
        d = {}
        for cls in self.descendants:
            d[cls.full_name()] = cls
        return d

    "Return the attribute names provided by this class only."

    class_attribute_names = NamespaceDict.keys

    def class_attributes(self):

        "Return class attributes provided by this class only."

        return dict(self)

    def all_class_attribute_names(self):

        "Return the attribute names provided by classes in this hierarchy."

        if self.all_classattr_names is None:
            self.all_class_attributes()
            self.all_classattr_names = self.all_classattr.keys()
        return self.all_classattr_names

    def all_class_attributes(self):

        "Return all class attributes, indicating the class which provides them."

        self.finalise_class_attributes()
        return self.all_classattr

    def finalise_class_attributes(self):

        "Make sure that the class attributes are fully defined."

        if self.all_classattr is None:
            self.all_classattr = {}
            clsattr = {}

            # Record provisional position information for attributes of this
            # class.

            for name in self.class_attributes().keys():
                clsattr[name] = set() # position not yet defined

            reversed_bases = self.bases[:]
            reversed_bases.reverse()

            # For the bases in reverse order, acquire class attribute details.

            for cls in reversed_bases:
                for name, attr in cls.all_class_attributes().items():
                    self.all_classattr[name] = attr

                    # Record previous attribute information.

                    if clsattr.has_key(name):
                        clsattr[name].add(attr.position)

            # Record class attributes provided by this class and its bases,
            # along with their positions.

            self.all_classattr.update(self.class_attributes())

            if clsattr:
                for i, name in enumerate(self._get_position_list(clsattr)):
                    self.all_classattr[name].position = i

        return self.all_classattr

    def instance_attribute_names(self):

        "Return the instance attribute names provided by the class."

        if self.all_instattr_names is None:
            self.instance_attributes()
        return self.all_instattr_names

    def instance_attributes(self):

        "Return instance-only attributes for instances of this class."

        self.finalise_instance_attributes()
        return self.all_instattr

    def finalise_instance_attributes(self):

        "Make sure that the instance attributes are fully defined."

        # Cache the attributes by converting the positioned attributes into a
        # dictionary.

        if self.all_instattr is None:
            self.all_instattr = self._get_attributes()
            self.all_instattr_names = self.all_instattr.keys()

        return self.all_instattr

    def _get_attributes(self):

        """
        Return a dictionary mapping names to Attr instances incorporating
        information about their positions in the final instance structure.
        """

        instattr = {}

        # Record provisional position information for attributes of this
        # instance.

        for name in self.instattr:
            instattr[name] = set() # position not yet defined

        reversed_bases = self.bases[:]
        reversed_bases.reverse()

        # For the bases in reverse order, acquire instance attribute
        # details.

        for cls in reversed_bases:
            for name, attr in cls.instance_attributes().items():

                # Record previous attribute information.

                if instattr.has_key(name):
                    instattr[name].add(attr.position)
                else:
                    instattr[name] = set([attr.position])

        # Build the dictionary of attributes using the existing positions known
        # for each name.

        d = {}
        for i, name in enumerate(self._get_position_list(instattr)):
            d[name] = Attr(i, Instance(), name)
        return d

    def _get_position_list(self, positions):

        """
        Return a list of attribute names for the given 'positions' mapping from
        names to positions, indicating the positions of the attributes in the
        final instance structure.
        """

        position_items = positions.items()
        namearray = [None] * len(position_items)

        # Get the positions in ascending order of list size, with lists
        # of the same size ordered according to their smallest position
        # value.

        position_items.sort(self._cmp_positions)

        # Get the names in position order.

        held = []

        for name, pos in position_items:
            pos = list(pos)
            pos.sort()
            if pos and pos[0] < len(namearray) and namearray[pos[0]] is None:
                namearray[pos[0]] = name
            else:
                if pos:
                    self.relocated.add(name)
                held.append((name, pos))

        for i, attr in enumerate(namearray):
            if attr is None:
                name, pos = held.pop()
                namearray[i] = name

        return namearray

    def _cmp_positions(self, a, b):

        "Compare name plus position list operands 'a' and 'b'."

        name_a, list_a = a
        name_b, list_b = b
        if len(list_a) < len(list_b):
            return -1
        elif len(list_a) > len(list_b):
            return 1
        elif not list_a:
            return 0
        else:
            return cmp(min(list_a), min(list_b))

    def all_attribute_names(self):

        """
        Return the names of all attributes provided by instances of this class.
        """

        self.allattr_names = self.allattr_names or self.all_attributes().keys()
        return self.allattr_names

    def all_attributes(self):

        """
        Return all attributes for an instance, indicating either the class which
        provides them or that the instance itself provides them.
        """

        if self.allattr is None:
            self.allattr = {}
            self.allattr.update(self.all_class_attributes())
            for name, attr in self.instance_attributes().items():
                if self.allattr.has_key(name):
                    print "Instance attribute %r in %r overrides class attribute." % (name, self)
                self.allattr[name] = attr
        return self.allattr

class Function(NamespaceDict, Naming, Constant):

    "An inspected function."

    def __init__(self, name, parent, argnames, defaults, has_star, has_dstar,
        dynamic_def=0, module=None, node=None):

        """
        Initialise the function with the given 'name', 'parent', list of
        'argnames', list of 'defaults', the 'has_star' flag (indicating the
        presence of a * parameter), the 'has_dstar' flag (indicating the
        presence of a ** parameter), optional 'module', and optional AST 'node'.
        """

        NamespaceDict.__init__(self, module)
        self.name = name
        self.parent = parent
        self.argnames = argnames
        self.defaults = defaults
        self.has_star = has_star
        self.has_dstar = has_dstar
        self.dynamic_def = dynamic_def
        self.astnode = node
        node._def = self

        # Initialise the positional names.

        self.positional_names = self.argnames[:]
        if has_dstar:
            self.dstar_name = self.positional_names[-1]
            del self.positional_names[-1]
        if has_star:
            self.star_name = self.positional_names[-1]
            del self.positional_names[-1]

        # Initialise default storage.
        # NOTE: This must be initialised separately due to the reliance on node
        # NOTE: visiting.

        self.default_attrs = []

        # Initialise attribute usage.

        for arg in argnames:

            # Define attribute users.

            if node is not None:
                self._define_attribute_user_for_name(node, arg)

            # Or just record the usage.

            else:
                self.attributes_used[-1][arg] = set()

        # Caches.

        self.localnames = None # cache for locals

        # Add parameters to the namespace.

        self._add_parameters(argnames)

        # Image generation details.

        self.dynamic = None
        self.location = None
        self.code_location = None
        self.code_body_location = None

        # Program-related details.

        self.blocks = None
        self.body_block = None

        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

    def _add_parameters(self, argnames):
        for name in argnames:
            if isinstance(name, tuple):
                self._add_parameters(name)
            else:
                self.set(name, Instance())

    def __repr__(self):
        if self.location is not None:
            return "Function(%r, %s, %r, location=%r, code_location=%r)" % (
                self.name, shortrepr(self.parent), self.argnames, self.location, self.code_location
                )
        else:
            return "Function(%r, %s, %r)" % (
                self.name, shortrepr(self.parent), self.argnames
                )

    def __shortrepr__(self):
        return "Function(%r, %s)" % (
            self.name, shortrepr(self.parent)
            )

    def get_body_block(self):
        return self.body_block

    # Defaults-related methods.

    def store_default(self, attr_or_value):

        """
        Reserve space for defaults, set outside the function, potentially on a
        dynamic basis, using the 'attr_or_value'.
        """

        attr = Attr(None, self, None)
        self._set_using_attr(attr, attr_or_value)
        self.default_attrs.append(attr)

    def make_dynamic(self):

        "Return whether this function must be handled using a dynamic object."

        if self.dynamic is None:
            for attr in self.default_attrs:
                if not attr.is_strict_constant() and self.dynamic_def:
                    self.dynamic = 1
                    self._make_dynamic()
                    break
            else:
                self.dynamic = 0

        return self.dynamic

    is_dynamic = make_dynamic

    def _make_dynamic(self):

        "Where functions have dynamic defaults, add a context argument."

        name = "<context>"
        self.argnames.insert(0, name)
        self.positional_names.insert(0, name)
        self.set(name, Instance())

    # Namespace-related methods.

    def make_global(self, name):

        "Declare 'name' as a global in the current namespace."

        if name not in self.argnames and not self.has_key(name):
            self.globals.add(name)
            return 1
        else:
            return 0

    def parameters(self):

        """
        Return a dictionary mapping parameter names to their position in the
        parameter list.
        """

        parameters = {}
        for i, name in enumerate(self.argnames):
            parameters[name] = i
        return parameters

    def all_locals(self):

        "Return a dictionary mapping names to local and parameter details."

        return dict(self)

    def locals(self):

        "Return a dictionary mapping names to local details."

        if self.localnames is None:
            self.localnames = {}
            self.localnames.update(self.all_locals())
            for name in self.argnames:
                del self.localnames[name]
        return self.localnames

    def is_method(self):

        """
        Return whether this function is a method explicitly defined in a class.
        """

        return isinstance(self.parent, Class)

    def is_relocated(self, name):

        """
        Determine whether the given attribute 'name' is relocated for instances
        having this function as a method.
        """

        for cls in self.parent.descendants:
            if name in cls.relocated:
                return 1
        return 0

    def finalise_attributes(self):

        """
        Make sure all attributes (locals) are fully defined. Note that locals
        are not attributes in the sense of class, module or instance attributes.
        Defaults are also finalised by this method.
        """

        if self.finalised:
            return

        # Defaults.

        for i, default in enumerate(self.default_attrs):
            default.position = i

        # Locals.

        i = None
        for i, name in enumerate(self.argnames):
            self[name].position = i

        if i is not None:
            nparams = i + 1
        else:
            nparams = 0

        i = None
        for i, attr in enumerate(self.locals().values()):
            attr.position = i + nparams

        if i is not None:
            nothers = i + 1
        else:
            nothers = 0

        self.local_usage = nothers
        self.all_local_usage = nparams + nothers
        self.finalised = 1

    def as_instantiator(self):

        "Make an instantiator function from a method, keeping all arguments."

        function = Function(self.parent.name, self.parent.parent, self.argnames, self.defaults,
            self.has_star, self.has_dstar, self.dynamic_def, self.module, self.astnode)
        function.default_attrs = self.default_attrs
        return function

class UnresolvedName(NamespaceDict, Constant):

    "A module, class or function which was mentioned but could not be imported."

    def __init__(self, name, parent_name, module=None):
        NamespaceDict.__init__(self, module)
        self.name = name
        self.parent_name = parent_name
        self.parent = None

        self.descendants = set()

    def add_descendant(self, cls):
        self.descendants.add(cls)

    def all_class_attributes(self):
        return {}

    def instance_attributes(self):
        return {}

    def __repr__(self):
        return "UnresolvedName(%r, %r)" % (self.name, self.parent_name)

    __shortrepr__ = __repr__

    def full_name(self):
        if self.name is not None:
            return self.parent_name + "." + self.name
        else:
            return self.parent_name

class Module(NamespaceDict, Constant):

    "An inspected module's core details."

    def __init__(self, name, importer):
        NamespaceDict.__init__(self, self)
        self.name = name
        self.importer = importer
        self.parent = None

        # Original location details.

        self.astnode = None

        # Complete lists of classes and functions.

        self.all_objects = set()

        # Keyword records.

        self.keyword_names = set()

        # Image generation details.

        self.location = None
        self.code_location = None

        # Program-related details.

        self.blocks = None
        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

    def full_name(self):
        return self.name

    def __repr__(self):
        if self.location is not None:
            return "Module(%r, location=%r)" % (self.name, self.location)
        else:
            return "Module(%r)" % self.name

    def __shortrepr__(self):
        return "Module(%r)" % self.name

    # Attribute methods.

    "Return the module attribute names provided by the module."

    module_attribute_names = NamespaceDict.keys

    def module_attributes(self):

        "Return a dictionary mapping names to module attributes."

        return dict(self)

# vim: tabstop=4 expandtab shiftwidth=4
